# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Text datasets."""

from tensorflow_datasets.summarization.aeslc import Aeslc
from tensorflow_datasets.summarization.big_patent import BigPatent
from tensorflow_datasets.summarization.billsum import Billsum
from tensorflow_datasets.summarization.booksum import Booksum
from tensorflow_datasets.summarization.cnn_dailymail import CnnDailymail
from tensorflow_datasets.summarization.covid19sum import Covid19sum
from tensorflow_datasets.summarization.gigaword import Gigaword
from tensorflow_datasets.summarization.gov_report import GovReport
from tensorflow_datasets.summarization.media_sum import MediaSum
from tensorflow_datasets.summarization.multi_news import MultiNews
from tensorflow_datasets.summarization.newsroom import Newsroom
from tensorflow_datasets.summarization.opinion_abstracts import OpinionAbstracts
from tensorflow_datasets.summarization.opinosis import Opinosis
from tensorflow_datasets.summarization.reddit import Reddit
from tensorflow_datasets.summarization.reddit_tifu import RedditTifu
from tensorflow_datasets.summarization.samsum import Samsum
from tensorflow_datasets.summarization.scientific_papers import ScientificPapers
from tensorflow_datasets.summarization.summscreen import Summscreen
from tensorflow_datasets.summarization.wikihow import Wikihow
from tensorflow_datasets.summarization.xsum import Xsum
