# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gsm8k dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.gsm8k import gsm8k


class Gsm8kTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gsm8k dataset."""

  DATASET_CLASS = gsm8k.Gsm8k
  SPLITS = {
      'train': 1,
      'test': 1,
      'train_socratic': 1,
      'test_socratic': 1,
  }
  DL_EXTRACT_RESULT = {
      k: f'{k}.jsonl'
      for k in ['train', 'test', 'train_socratic', 'test_socratic']
  }


if __name__ == '__main__':
  tfds.testing.test_main()
