# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLI Fixtures."""

import argparse
from unittest import mock

import pytest


@pytest.fixture(scope='session', autouse=True)
def _mock_argparse_flags():
  """Mock absl argparse_flag with argparse."""
  # absl flags are globally registered during pytest tests collection.
  # This creates conflict if the same flag is defined outside the CLI (in
  # another test):
  # `flags.DEFINE_string('data_dir')` with `parser.add_argument('--data_dir')`
  # We patch argparse_flags during test, so absl flags are ignored.
  with mock.patch(
      'absl.flags.argparse_flags.ArgumentParser', argparse.ArgumentParser
  ):
    yield
