# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: waymo_dataset.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13waymo_dataset.proto\x12\x12waymo.open_dataset"\x8f\x06\n\x05Label\x12*\n\x03\x62ox\x18\x01'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Label.Box\x12\x34\n\x08metadata\x18\x02'
    b' \x01(\x0b\x32".waymo.open_dataset.Label.Metadata\x12,\n\x04type\x18\x03'
    b' \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\n\n\x02id\x18\x04'
    b' \x01(\t\x12M\n\x1a\x64\x65tection_difficulty_level\x18\x05'
    b' \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12L\n\x19tracking_difficulty_level\x18\x06'
    b' \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12\x1f\n\x17num_lidar_points_in_box\x18\x07'
    b' \x01(\x05\x1a\xbf\x01\n\x03\x42ox\x12\x10\n\x08\x63\x65nter_x\x18\x01'
    b' \x01(\x01\x12\x10\n\x08\x63\x65nter_y\x18\x02'
    b' \x01(\x01\x12\x10\n\x08\x63\x65nter_z\x18\x03'
    b' \x01(\x01\x12\x0e\n\x06length\x18\x05 \x01(\x01\x12\r\n\x05width\x18\x04'
    b' \x01(\x01\x12\x0e\n\x06height\x18\x06'
    b' \x01(\x01\x12\x0f\n\x07heading\x18\x07'
    b' \x01(\x01"B\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x0b\n\x07TYPE_3D\x10\x01\x12\x0b\n\x07TYPE_2D\x10\x02\x12\x0e\n\nTYPE_AA_2D\x10\x03\x1aN\n\x08Metadata\x12\x0f\n\x07speed_x\x18\x01'
    b' \x01(\x01\x12\x0f\n\x07speed_y\x18\x02'
    b' \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_x\x18\x03'
    b' \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_y\x18\x04'
    b' \x01(\x01"`\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x10\n\x0cTYPE_VEHICLE\x10\x01\x12\x13\n\x0fTYPE_PEDESTRIAN\x10\x02\x12\r\n\tTYPE_SIGN\x10\x03\x12\x10\n\x0cTYPE_CYCLIST\x10\x04"8\n\x0f\x44ifficultyLevel\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07LEVEL_1\x10\x01\x12\x0b\n\x07LEVEL_2\x10\x02"2\n\x0ePolygon2dProto\x12\t\n\x01x\x18\x01'
    b' \x03(\x01\x12\t\n\x01y\x18\x02 \x03(\x01\x12\n\n\x02id\x18\x03'
    b' \x01(\t"\x1b\n\x0bMatrixShape\x12\x0c\n\x04\x64ims\x18\x01'
    b' \x03(\x05"O\n\x0bMatrixFloat\x12\x10\n\x04\x64\x61ta\x18\x01'
    b' \x03(\x02\x42\x02\x10\x01\x12.\n\x05shape\x18\x02'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape"O\n\x0bMatrixInt32\x12\x10\n\x04\x64\x61ta\x18\x01'
    b' \x03(\x05\x42\x02\x10\x01\x12.\n\x05shape\x18\x02'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape"l\n\nCameraName"^\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x46RONT\x10\x01\x12\x0e\n\nFRONT_LEFT\x10\x02\x12\x0f\n\x0b\x46RONT_RIGHT\x10\x03\x12\r\n\tSIDE_LEFT\x10\x04\x12\x0e\n\nSIDE_RIGHT\x10\x05"]\n\tLaserName"P\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03TOP\x10\x01\x12\t\n\x05\x46RONT\x10\x02\x12\r\n\tSIDE_LEFT\x10\x03\x12\x0e\n\nSIDE_RIGHT\x10\x04\x12\x08\n\x04REAR\x10\x05"\x1e\n\tTransform\x12\x11\n\ttransform\x18\x01'
    b' \x03(\x01"X\n\x08Velocity\x12\x0b\n\x03v_x\x18\x01'
    b' \x01(\x02\x12\x0b\n\x03v_y\x18\x02 \x01(\x02\x12\x0b\n\x03v_z\x18\x03'
    b' \x01(\x02\x12\x0b\n\x03w_x\x18\x04 \x01(\x01\x12\x0b\n\x03w_y\x18\x05'
    b' \x01(\x01\x12\x0b\n\x03w_z\x18\x06'
    b' \x01(\x01"\xa3\x03\n\x11\x43\x61meraCalibration\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\x11\n\tintrinsic\x18\x02'
    b' \x03(\x01\x12\x30\n\textrinsic\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12\r\n\x05width\x18\x04'
    b' \x01(\x05\x12\x0e\n\x06height\x18\x05'
    b' \x01(\x05\x12g\n\x19rolling_shutter_direction\x18\x06'
    b' \x01(\x0e\x32\x44.waymo.open_dataset.CameraCalibration.RollingShutterReadOutDirection"\x8d\x01\n\x1eRollingShutterReadOutDirection\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x11\n\rTOP_TO_BOTTOM\x10\x01\x12\x11\n\rLEFT_TO_RIGHT\x10\x02\x12\x11\n\rBOTTOM_TO_TOP\x10\x03\x12\x11\n\rRIGHT_TO_LEFT\x10\x04\x12\x12\n\x0eGLOBAL_SHUTTER\x10\x05"\xcd\x01\n\x10LaserCalibration\x12\x30\n\x04name\x18\x01'
    b' \x01(\x0e\x32".waymo.open_dataset.LaserName.Name\x12\x19\n\x11\x62\x65\x61m_inclinations\x18\x02'
    b' \x03(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_min\x18\x03'
    b' \x01(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_max\x18\x04'
    b' \x01(\x01\x12\x30\n\textrinsic\x18\x05'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform"\xf6\x03\n\x07\x43ontext\x12\x0c\n\x04name\x18\x01'
    b' \x01(\t\x12\x42\n\x13\x63\x61mera_calibrations\x18\x02'
    b' \x03(\x0b\x32%.waymo.open_dataset.CameraCalibration\x12@\n\x12laser_calibrations\x18\x03'
    b' \x03(\x0b\x32$.waymo.open_dataset.LaserCalibration\x12\x30\n\x05stats\x18\x04'
    b' \x01(\x0b\x32!.waymo.open_dataset.Context.Stats\x1a\xa4\x02\n\x05Stats\x12J\n\x13laser_object_counts\x18\x01'
    b' \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12K\n\x14\x63\x61mera_object_counts\x18\x05'
    b' \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12\x13\n\x0btime_of_day\x18\x02'
    b' \x01(\t\x12\x10\n\x08location\x18\x03'
    b' \x01(\t\x12\x0f\n\x07weather\x18\x04'
    b' \x01(\t\x1aJ\n\x0bObjectCount\x12,\n\x04type\x18\x01'
    b' \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\r\n\x05\x63ount\x18\x02'
    b' \x01(\x05"\xb1\x01\n\nRangeImage\x12\x1e\n\x16range_image_compressed\x18\x02'
    b' \x01(\x0c\x12$\n\x1c\x63\x61mera_projection_compressed\x18\x03'
    b' \x01(\x0c\x12#\n\x1brange_image_pose_compressed\x18\x04'
    b' \x01(\x0c\x12\x38\n\x0brange_image\x18\x01'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixFloatB\x02\x18\x01"\x94\x02\n\x0b\x43\x61meraImage\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\r\n\x05image\x18\x02'
    b' \x01(\x0c\x12+\n\x04pose\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12.\n\x08velocity\x18\x04'
    b' \x01(\x0b\x32\x1c.waymo.open_dataset.Velocity\x12\x16\n\x0epose_timestamp\x18\x05'
    b' \x01(\x01\x12\x0f\n\x07shutter\x18\x06'
    b' \x01(\x01\x12\x1b\n\x13\x63\x61mera_trigger_time\x18\x07 \x01(\x01\x12'
    b' \n\x18\x63\x61mera_readout_done_time\x18\x08'
    b' \x01(\x01"l\n\x0c\x43\x61meraLabels\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12)\n\x06labels\x18\x02'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Label"\xa1\x01\n\x05Laser\x12\x30\n\x04name\x18\x01'
    b' \x01(\x0e\x32".waymo.open_dataset.LaserName.Name\x12\x32\n\nri_return1\x18\x02'
    b' \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage\x12\x32\n\nri_return2\x18\x03'
    b' \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage"\xc0\x03\n\x05\x46rame\x12,\n\x07\x63ontext\x18\x01'
    b' \x01(\x0b\x32\x1b.waymo.open_dataset.Context\x12\x18\n\x10timestamp_micros\x18\x02'
    b' \x01(\x03\x12+\n\x04pose\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12/\n\x06images\x18\x04'
    b' \x03(\x0b\x32\x1f.waymo.open_dataset.CameraImage\x12)\n\x06lasers\x18\x05'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Laser\x12/\n\x0claser_labels\x18\x06'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Label\x12@\n\x16projected_lidar_labels\x18\t'
    b' \x03(\x0b\x32'
    b' .waymo.open_dataset.CameraLabels\x12\x37\n\rcamera_labels\x18\x08'
    b' \x03(\x0b\x32'
    b' .waymo.open_dataset.CameraLabels\x12:\n\x0eno_label_zones\x18\x07'
    b' \x03(\x0b\x32".waymo.open_dataset.Polygon2dProto'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'waymo_dataset_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _MATRIXFLOAT.fields_by_name['data']._options = None
  _MATRIXFLOAT.fields_by_name['data']._serialized_options = b'\020\001'
  _MATRIXINT32.fields_by_name['data']._options = None
  _MATRIXINT32.fields_by_name['data']._serialized_options = b'\020\001'
  _RANGEIMAGE.fields_by_name['range_image']._options = None
  _RANGEIMAGE.fields_by_name['range_image']._serialized_options = b'\030\001'
  _LABEL._serialized_start = 44
  _LABEL._serialized_end = 827
  _LABEL_BOX._serialized_start = 400
  _LABEL_BOX._serialized_end = 591
  _LABEL_BOX_TYPE._serialized_start = 525
  _LABEL_BOX_TYPE._serialized_end = 591
  _LABEL_METADATA._serialized_start = 593
  _LABEL_METADATA._serialized_end = 671
  _LABEL_TYPE._serialized_start = 673
  _LABEL_TYPE._serialized_end = 769
  _LABEL_DIFFICULTYLEVEL._serialized_start = 771
  _LABEL_DIFFICULTYLEVEL._serialized_end = 827
  _POLYGON2DPROTO._serialized_start = 829
  _POLYGON2DPROTO._serialized_end = 879
  _MATRIXSHAPE._serialized_start = 881
  _MATRIXSHAPE._serialized_end = 908
  _MATRIXFLOAT._serialized_start = 910
  _MATRIXFLOAT._serialized_end = 989
  _MATRIXINT32._serialized_start = 991
  _MATRIXINT32._serialized_end = 1070
  _CAMERANAME._serialized_start = 1072
  _CAMERANAME._serialized_end = 1180
  _CAMERANAME_NAME._serialized_start = 1086
  _CAMERANAME_NAME._serialized_end = 1180
  _LASERNAME._serialized_start = 1182
  _LASERNAME._serialized_end = 1275
  _LASERNAME_NAME._serialized_start = 1195
  _LASERNAME_NAME._serialized_end = 1275
  _TRANSFORM._serialized_start = 1277
  _TRANSFORM._serialized_end = 1307
  _VELOCITY._serialized_start = 1309
  _VELOCITY._serialized_end = 1397
  _CAMERACALIBRATION._serialized_start = 1400
  _CAMERACALIBRATION._serialized_end = 1819
  _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION._serialized_start = 1678
  _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION._serialized_end = 1819
  _LASERCALIBRATION._serialized_start = 1822
  _LASERCALIBRATION._serialized_end = 2027
  _CONTEXT._serialized_start = 2030
  _CONTEXT._serialized_end = 2532
  _CONTEXT_STATS._serialized_start = 2240
  _CONTEXT_STATS._serialized_end = 2532
  _CONTEXT_STATS_OBJECTCOUNT._serialized_start = 2458
  _CONTEXT_STATS_OBJECTCOUNT._serialized_end = 2532
  _RANGEIMAGE._serialized_start = 2535
  _RANGEIMAGE._serialized_end = 2712
  _CAMERAIMAGE._serialized_start = 2715
  _CAMERAIMAGE._serialized_end = 2991
  _CAMERALABELS._serialized_start = 2993
  _CAMERALABELS._serialized_end = 3101
  _LASER._serialized_start = 3104
  _LASER._serialized_end = 3265
  _FRAME._serialized_start = 3268
  _FRAME._serialized_end = 3716
# @@protoc_insertion_point(module_scope)
